import os
from enum import Enum
import httpx
from typing import List, Dict, Any, Optional, Union, Mapping, Literal, TypeVar, cast
from typing_extensions import TypedDict
from pydantic import BaseModel
from .exceptions import (
    APIStatusError,
    BadRequestError,
    AuthenticationError,
    PermissionDeniedError,
    NotFoundError,
    ConflictError,
    UnprocessableEntityError,
    RateLimitError,
    InternalServerError,
)


class RubeusCacheType(Enum):
    SEMANTIC = "semantic"
    SIMPLE = "simple"


RubeusCacheLiteral = Literal["semantic", "simple"]


ResponseT = TypeVar("ResponseT", bound="RubeusResponse")


class ProviderTypes(str, Enum):
    """_summary_

    Args:
        Enum (_type_): _description_

    Returns:
        _type_: _description_
    """

    OPENAI = "openai"
    COHERE = "cohere"
    ANTHROPIC = "anthropic"
    AZURE_OPENAI = "azure-openai"
    HUGGING_FACE = "huggingface"


ProviderTypesLiteral = Literal[
    "openai", "cohere", "anthropic", "azure-openai", "huggingface"
]


class RubeusModes(str, Enum):
    """_summary_

    Args:
        Enum (_type_): _description_
    """

    FALLBACK = "fallback"
    LOADBALANCE = "loadbalance"
    SINGLE = "single"
    PROXY = "proxy"


RubeusModesLiteral = Literal["fallback", "loadbalance", "single", "proxy"]


class RubeusApiPaths(Enum):
    CHAT_COMPLETION = "/v1/chatComplete"
    COMPLETION = "/v1/complete"


class Options(BaseModel):
    method: str
    url: str
    params: Optional[Mapping[str, str]] = None
    headers: Optional[Mapping[str, str]] = None
    max_retries: Optional[int] = None
    timeout: Optional[float] = None
    # stringified json
    data: Optional[Mapping[str, Any]] = None
    # json structure
    json_body: Optional[Mapping[str, Any]] = None


class Message(TypedDict):
    role: str
    content: str


class Function(BaseModel):
    name: str
    description: str
    parameters: str


class RetrySettings(BaseModel):
    attempts: int
    on_status_codes: list


class OverrideParams(TypedDict, total=False):
    provider: Optional[Union[ProviderTypes, ProviderTypesLiteral]]
    prompt: Optional[str]
    messages: Optional[List[Message]]
    temperature: Optional[float]
    max_tokens: Optional[int]
    max_retries: Optional[int]
    trace_id: Optional[str]
    cache_status: Optional[Union[RubeusCacheType, RubeusCacheLiteral]]
    cache: Optional[bool]
    metadata: Optional[Dict[str, Any]]
    weight: Optional[float]
    top_k: Optional[int]
    top_p: Optional[float]
    stop_sequences: Optional[List[str]]
    timeout: Union[float, None]
    retry_settings: Optional[RetrySettings]
    model: Optional[str]
    api_key: Optional[str]
    stream: Optional[bool]
    functions: Optional[List[Function]]
    function_call: Optional[Union[None, str, Function]]
    n: Optional[int]
    logprobs: Optional[int]
    echo: Optional[bool]
    stop: Optional[Union[str, List[str]]]
    presence_penalty: Optional[int]
    frequency_penalty: Optional[int]
    best_of: Optional[int]
    logit_bias: Optional[Mapping[str, int]]
    user: Optional[str]


class ProviderOptions(BaseModel):
    provider: Optional[str]
    apiKey: Optional[str]
    weight: Optional[float]
    override_params: Optional[OverrideParams]
    retry: Optional[RetrySettings]


class Config(BaseModel):
    mode: str
    options: List[ProviderOptions]


def remove_empty_values(
    data: Union[Dict[str, Any], Mapping[str, Any]]
) -> Dict[str, Any]:
    if isinstance(data, dict):
        cleaned_dict = {}
        for key, value in data.items():
            if value is not None and value != "":
                cleaned_value = remove_empty_values(value)
                if cleaned_value is not None and cleaned_value != "":
                    cleaned_dict[key] = cleaned_value
        return cleaned_dict
    elif isinstance(data, list):
        cleaned_list = []

        for item in data:  # type: ignore
            cleaned_item = remove_empty_values(item)
            if cleaned_item is not None and cleaned_item != "":
                cleaned_list.append(cleaned_item)
        return cleaned_list  # type: ignore
    else:
        return cast(dict, data)


# class DefaultHeaders(TypedDict. total=false)


class DefaultParams(BaseModel):
    provider: Optional[Union[ProviderTypes, ProviderTypesLiteral]] = None
    prompt: Optional[str] = None
    messages: Optional[List[Message]] = None
    temperature: Optional[float] = None
    max_tokens: Optional[int] = None
    max_retries: Optional[int] = None
    trace_id: Optional[str] = None
    cache_status: Optional[Union[RubeusCacheType, RubeusCacheLiteral]] = None
    cache: Optional[bool] = None
    metadata: Optional[Dict[str, Any]] = None
    weight: Optional[float] = None
    top_k: Optional[int] = None
    top_p: Optional[float] = None
    stop_sequences: Optional[List[str]] = None
    timeout: Union[float, None] = None
    retry_settings: Optional[RetrySettings] = None
    model: Optional[str] = None
    api_key: Optional[str] = None
    stream: Optional[bool] = False
    functions: Optional[List[Function]]
    function_call: Optional[Union[None, str, Function]]
    n: Optional[int]
    logprobs: Optional[int]
    echo: Optional[bool]
    stop: Optional[Union[str, List[str]]]
    presence_penalty: Optional[int]
    frequency_penalty: Optional[int]
    best_of: Optional[int]
    logit_bias: Optional[Dict[str, int]]
    user: Optional[str]

    def __init__(
        self,
        *,
        prompt: Optional[str] = None,
        messages: Optional[List[Message]] = None,
        provider: Optional[Union[ProviderTypes, ProviderTypesLiteral]] = None,
        model: Optional[str] = None,
        api_key: Optional[str] = None,
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = None,
        max_retries: Optional[int] = None,
        trace_id: Optional[str] = None,
        cache_status: Optional[Union[RubeusCacheType, RubeusCacheLiteral]] = None,
        cache: Optional[bool] = None,
        metadata: Optional[Dict[str, Any]] = None,
        weight: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        stop_sequences: Optional[List[str]] = None,
        stream: Optional[bool] = False,
        timeout: Union[float, None] = None,
        retry_settings: Optional[RetrySettings] = None,
        functions: Optional[List[Function]] = None,
        function_call: Optional[Union[None, str, Function]] = None,
        n: Optional[int] = None,
        logprobs: Optional[int] = None,
        echo: Optional[bool] = None,
        stop: Optional[Union[str, List[str]]] = None,
        presence_penalty: Optional[int] = None,
        frequency_penalty: Optional[int] = None,
        best_of: Optional[int] = None,
        logit_bias: Optional[Dict[str, int]] = None,
        user: Optional[str] = None,
    ):
        api_key = api_key or apikey_from_env(provider)
        super().__init__(
            prompt=prompt,
            messages=messages,
            provider=provider,
            model=model,
            api_key=api_key,
            temperature=temperature,
            max_tokens=max_tokens,
            max_retries=max_retries,
            trace_id=trace_id,
            cache_status=cache_status,
            cache=cache,
            metadata=metadata,
            weight=weight,
            top_k=top_k,
            top_p=top_p,
            stop_sequences=stop_sequences,
            stream=stream,
            timeout=timeout,
            retry_settings=retry_settings,
            functions=functions,
            function_call=function_call,
            n=n,
            logprobs=logprobs,
            echo=echo,
            stop=stop,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
            best_of=best_of,
            logit_bias=logit_bias,
            user=user,
        )


class LLMBase(BaseModel):
    """
    provider (Optional[ProviderTypes]): The LLM provider to be used for the Portkey
    integration.
        Eg: openai, anthropic etc.
        NOTE: Check the ProviderTypes to see the supported list of LLMs.
    model (str): The name of the language model to use (default: "gpt-3.5-turbo").
    temperature (float): The temperature parameter for text generation (default: 0.1).
    max_tokens (Optional[int]): The maximum number of tokens in the generated text.
    max_retries (int): The maximum number of retries for failed requests (default: 5).
    trace_id (Optional[str]): A unique identifier for tracing requests.
    cache_status (Optional[Union[RubeusCacheType,  RubeusCacheLiteral]]): The type of c
    ache to use (default: "").
        If cache_status is set, then cache is automatically set to True
    cache (Optional[bool]): Whether to use caching (default: False).
    metadata (Optional[Dict[str, Any]]): Metadata associated with the request
    (default: {}).
    weight (Optional[float]): The weight of the LLM in the ensemble (default: 1.0).
    """

    prompt: Optional[str] = None
    messages: Optional[List[Message]] = None
    provider: Union[ProviderTypes, ProviderTypesLiteral]
    model: str
    api_key: str
    temperature: Optional[float] = None
    max_tokens: Optional[int] = None
    max_retries: Optional[int] = None
    trace_id: Optional[str] = None
    cache_status: Optional[Union[RubeusCacheType, RubeusCacheLiteral]] = None
    cache: Optional[bool] = None
    metadata: Optional[Dict[str, Any]] = None
    weight: Optional[float] = None
    top_k: Optional[int] = None
    top_p: Optional[float] = None
    stop_sequences: Optional[List[str]] = None
    timeout: Union[float, None] = None
    retry_settings: Optional[RetrySettings] = None
    functions: Optional[List[Function]]
    function_call: Optional[Union[None, str, Function]]
    n: Optional[int]
    logprobs: Optional[int]
    echo: Optional[bool]
    stop: Optional[Union[str, List[str]]]
    presence_penalty: Optional[int]
    frequency_penalty: Optional[int]
    best_of: Optional[int]
    logit_bias: Optional[Dict[str, int]]
    user: Optional[str]

    # NOTE: We do not support streaming in over-ride params.
    def __init__(
        self,
        *,
        prompt: Optional[str] = None,
        messages: Optional[List[Message]] = None,
        provider: Union[ProviderTypes, ProviderTypesLiteral],
        model: str,
        api_key: Optional[str] = None,
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = None,
        max_retries: Optional[int] = None,
        trace_id: Optional[str] = None,
        cache_status: Optional[RubeusCacheType] = None,
        cache: Optional[bool] = None,
        metadata: Optional[Dict[str, Any]] = None,
        weight: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        stop_sequences: Optional[List[str]] = None,
        timeout: Union[float, None] = None,
        retry_settings: Optional[RetrySettings] = None,
        functions: Optional[List[Function]] = None,
        function_call: Optional[Union[None, str, Function]] = None,
        n: Optional[int] = None,
        logprobs: Optional[int] = None,
        echo: Optional[bool] = None,
        stop: Optional[Union[str, List[str]]] = None,
        presence_penalty: Optional[int] = None,
        frequency_penalty: Optional[int] = None,
        best_of: Optional[int] = None,
        logit_bias: Optional[Dict[str, int]] = None,
        user: Optional[str] = None,
    ):
        api_key = api_key or apikey_from_env(provider)
        super().__init__(
            prompt=prompt,
            messages=messages,
            provider=provider,
            model=model,
            api_key=api_key,
            temperature=temperature,
            max_tokens=max_tokens,
            max_retries=max_retries,
            trace_id=trace_id,
            cache_status=cache_status,
            cache=cache,
            metadata=metadata,
            weight=weight,
            top_k=top_k,
            top_p=top_p,
            stop_sequences=stop_sequences,
            timeout=timeout,
            retry_settings=retry_settings,
            functions=functions,
            function_call=function_call,
            n=n,
            logprobs=logprobs,
            echo=echo,
            stop=stop,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
            best_of=best_of,
            logit_bias=logit_bias,
            user=user,
        )

    def override_params(self) -> OverrideParams:
        return {
            "prompt": self.prompt,
            "messages": self.messages,
            "provider": self.provider,
            "model": self.model,
            "api_key": self.api_key,
            "temperature": self.temperature,
            "max_tokens": self.max_tokens,
            "max_retries": self.max_retries,
            "trace_id": self.trace_id,
            "cache_status": self.cache_status,
            "cache": self.cache,
            "metadata": self.metadata,
            "weight": self.weight,
            "top_k": self.top_k,
            "top_p": self.top_p,
            "stop_sequences": self.stop_sequences,
            "timeout": self.timeout,
            "retry_settings": self.retry_settings,
            "functions": self.functions,
            "function_call": self.function_call,
            "n": self.n,
            "logprobs": self.logprobs,
            "echo": self.echo,
            "stop": self.stop,
            "presence_penalty": self.presence_penalty,
            "frequency_penalty": self.frequency_penalty,
            "best_of": self.best_of,
            "logit_bias": self.logit_bias,
            "user": self.user,
        }


class Body(LLMBase):
    pass


class Params(LLMBase):
    pass


class RequestData(BaseModel):
    config: Config
    params: Params


class RubeusResponse(BaseModel):
    model: str
    choices: List[Any]
    raw_body: Dict[str, Any]


def apikey_from_env(provider: Union[ProviderTypes, ProviderTypesLiteral, None]) -> str:
    if provider == ProviderTypes.OPENAI:
        return os.environ.get("OPENAI_API_KEY", "")
    elif provider == ProviderTypes.COHERE:
        return os.environ.get("COHERE_API_KEY", "")
    elif provider == ProviderTypes.AZURE_OPENAI:
        return os.environ.get("AZURE_OPENAI_API_KEY", "")
    elif provider == ProviderTypes.HUGGING_FACE:
        return os.environ.get("HUGGINGFACE_API_KEY", "")
    return ""


def make_status_error(
    err_msg: str,
    *,
    body: object,
    request: httpx.Request,
    response: httpx.Response,
) -> APIStatusError:
    if response.status_code == 400:
        return BadRequestError(err_msg, request=request, response=response, body=body)
    if response.status_code == 401:
        return AuthenticationError(
            err_msg, request=request, response=response, body=body
        )
    if response.status_code == 403:
        return PermissionDeniedError(
            err_msg, request=request, response=response, body=body
        )
    if response.status_code == 404:
        return NotFoundError(err_msg, request=request, response=response, body=body)
    if response.status_code == 409:
        return ConflictError(err_msg, request=request, response=response, body=body)
    if response.status_code == 422:
        return UnprocessableEntityError(
            err_msg, request=request, response=response, body=body
        )
    if response.status_code == 429:
        return RateLimitError(err_msg, request=request, response=response, body=body)
    if response.status_code >= 500:
        return InternalServerError(
            err_msg, request=request, response=response, body=body
        )
    return APIStatusError(err_msg, request=request, response=response, body=body)



from __future__ import annotations

import json
from typing import Any, Iterator, Generic, cast, Union

import httpx

from .utils import ResponseT, RubeusResponse, make_status_error


class ServerSentEvent:
    def __init__(
        self,
        *,
        event: Union[str, None] = None,
        data: Union[str, None] = None,
        id: Union[str, None] = None,
        retry: Union[int, None] = None,
    ) -> None:
        if data is None:
            data = ""

        self._id = id
        self._data = data
        self._event = event or None
        self._retry = retry

    @property
    def event(self) -> Union[str, None]:
        return self._event

    @property
    def id(self) -> Union[str, None]:
        return self._id

    @property
    def retry(self) -> Union[int, None]:
        return self._retry

    @property
    def data(self) -> str:
        return self._data

    def json(self) -> Any:
        return (
            {"model": "", "choices": [{}]}
            if self.data == "[DONE]"
            else json.loads(self.data)
        )

    def __repr__(self) -> str:
        return f"ServerSentEvent(event={self.event}, data={self.data}, id={self.id},\
            retry={self.retry})"


class SSEDecoder:
    _data: list[str]
    _event: Union[str, None]
    _retry: Union[int, None]
    _last_event_id: Union[str, None]

    def __init__(self) -> None:
        self._event = None
        self._data = []
        self._last_event_id = None
        self._retry = None

    def iter(self, iterator: Iterator[str]) -> Iterator[ServerSentEvent]:
        """Given an iterator that yields lines, iterate over it & yield every
        event encountered
        """
        for line in iterator:
            line = line.rstrip("\n")
            sse = self.decode(line)
            if sse is not None:
                yield sse

    def decode(self, line: str) -> Union[ServerSentEvent, None]:
        # See: https://html.spec.whatwg.org/multipage/server-sent-events.html#event-stream-interpretation  # noqa: E501

        if not line:
            if (
                not self._event
                and not self._data
                and not self._last_event_id
                and self._retry is None
            ):
                return None

            sse = ServerSentEvent(
                event=self._event,
                data="\n".join(self._data),
                id=self._last_event_id,
                retry=self._retry,
            )

            # NOTE: as per the SSE spec, do not reset last_event_id.
            self._event = None
            self._data = []
            self._retry = None

            return sse

        if line.startswith(":"):
            return None

        fieldname, _, value = line.partition(":")

        if value.startswith(" "):
            value = value[1:]

        if fieldname == "event":
            self._event = value
        elif fieldname == "data":
            self._data.append(value)
        elif fieldname == "id":
            if "\0" in value:
                pass
            else:
                self._last_event_id = value
        elif fieldname == "retry":
            try:
                self._retry = int(value)
            except (TypeError, ValueError):
                pass
        else:
            pass  # Field is ignored.

        return None


class Stream(Generic[ResponseT]):
    """Provides the core interface to iterate over a synchronous stream response."""

    response: httpx.Response

    def __init__(
        self,
        *,
        response: httpx.Response,
    ) -> None:
        self.response = response
        self._decoder = SSEDecoder()
        self._iterator = self.__stream__()

    def __next__(self) -> ResponseT:
        return self._iterator.__next__()

    def __iter__(self) -> Iterator[ResponseT]:
        for item in self._iterator:
            yield item

    def _iter_events(self) -> Iterator[ServerSentEvent]:
        yield from self._decoder.iter(self.response.iter_lines())

    def __stream__(self) -> Iterator[ResponseT]:
        response = self.response
        for sse in self._iter_events():
            if sse.event is None:
                yield cast(
                    ResponseT,
                    RubeusResponse.construct(**sse.json(), raw_body=sse.json()),
                )

            if sse.event == "ping":
                continue

            if sse.event == "error":
                body = sse.data

                try:
                    body = sse.json()
                    err_msg = f"{body}"
                except Exception:
                    err_msg = sse.data or f"Error code: {response.status_code}"

                raise make_status_error(
                    err_msg,
                    body=body,
                    response=self.response,
                    request=self.response.request,
                )



"""main file"""
import argparse
from rubeus.version import VERSION


def main():
    """Main function"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--version",
        "-v",
        action="version",
        version=f"rubeus {VERSION}",
        help="Print version and exit.",
    )

    _ = parser.parse_args()


if __name__ == "__main__":
    main()



MISSING_API_KEY_ERROR_MESSAGE = """No API key found for Portkey.
Please set either the PORTKEY_API_KEY environment variable or \
pass the api_key prior to initialization of Portkey.
API keys can be found or created at Portkey Dashboard \

Here's how you get it:
1. Visit https://app.portkey.ai/
1. Click on your profile icon on the top left
2. From the dropdown menu, click on "Copy API Key"
"""


DEFAULT_MAX_RETRIES = 2
VERSION = "0.1.0"
DEFAULT_TIMEOUT = 60



from __future__ import annotations

import json
from types import TracebackType
from typing import (
    Dict,
    Any,
    Union,
    Mapping,
    cast,
    List,
    Optional,
    Type,
    overload,
    Literal,
)
import httpx
import platform
from .global_constants import DEFAULT_MAX_RETRIES
from .utils import (
    remove_empty_values,
    Body,
    Options,
    Config,
    ProviderOptions,
    RubeusResponse,
)
from .exceptions import (
    APIStatusError,
    APITimeoutError,
    APIConnectionError,
)
from rubeus.version import VERSION
from .utils import ResponseT, make_status_error
from .common_types import StreamT
from .streaming import Stream


class MissingStreamClassError(TypeError):
    def __init__(self) -> None:
        super().__init__(
            "The `stream` argument was set to `True` but the `stream_cls` argument was\
            not given",
        )


class APIClient:
    _client: httpx.Client
    _default_stream_cls: Union[type[Stream[Any]], None] = None

    def __init__(
        self,
        *,
        base_url: str,
        api_key: str,
        timeout: Union[float, None],
        max_retries: int = DEFAULT_MAX_RETRIES,
        custom_headers: Optional[Mapping[str, Any]] = None,
        custom_query: Optional[Mapping[str, object]],
        custom_params: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.api_key = api_key
        self.max_retries = max_retries
        self._custom_headers = custom_headers
        self._custom_query = custom_query
        self._custom_params = custom_params
        self._client = httpx.Client(
            base_url=base_url,
            timeout=timeout,
            headers={"Accept": "application/json"},
        )

    @property
    def custom_auth(self) -> Optional[httpx.Auth]:
        return None

    @overload
    def post(
        self,
        path: str,
        *,
        body: List[Body],
        mode: str,
        cast_to: Type[ResponseT],
        stream: Literal[True],
        stream_cls: type[StreamT],
    ) -> StreamT:
        ...

    @overload
    def post(
        self,
        path: str,
        *,
        body: List[Body],
        mode: str,
        cast_to: Type[ResponseT],
        stream: Literal[False],
        stream_cls: type[StreamT],
    ) -> ResponseT:
        ...

    @overload
    def post(
        self,
        path: str,
        *,
        body: List[Body],
        mode: str,
        cast_to: Type[ResponseT],
        stream: bool,
        stream_cls: type[StreamT],
    ) -> Union[ResponseT, StreamT]:
        ...

    def post(
        self,
        path: str,
        *,
        body: List[Body],
        mode: str,
        cast_to: Type[ResponseT],
        stream: bool,
        stream_cls: type[StreamT],
    ) -> Union[ResponseT, StreamT]:
        body = cast(List[Body], body)
        opts = self._construct(method="post", url=path, body=body, mode=mode)
        res = self._request(
            options=opts,
            stream=stream,
            cast_to=cast_to,
            stream_cls=stream_cls,
        )
        return res

    def _construct(
        self, *, method: str, url: str, body: List[Body], mode: str
    ) -> Options:
        opts = Options.construct()
        opts.method = method
        opts.url = url
        json_body = {
            "config": self._config(mode, body).dict(),
            "params": self._custom_params,
        }
        opts.json_body = remove_empty_values(json_body)
        opts.headers = (
            remove_empty_values(self._custom_headers) if self._custom_headers else None
        )
        return opts

    def _config(self, mode: str, body: List[Body]) -> Config:
        config = Config(mode=mode, options=[])
        for i in body:
            item = i.dict()
            override_params = i.override_params()
            options = ProviderOptions(
                provider=item.get("provider"),
                apiKey=item.get("api_key"),
                weight=item.get("weight"),
                retry=item.get("retry"),
                override_params=override_params,
            )
            config.options.append(options)
        return config

    @property
    def _default_headers(self) -> Mapping[str, str]:
        return {
            "Content-Type": "application/json",
            "x-portkey-api-key": self.api_key,
            "x-rubeus-package-version": VERSION,
            "x-rubeus-runtime": platform.python_implementation(),
            "x-rubeus-runtime-version": platform.python_version(),
        }

    def _build_headers(self, options: Options) -> httpx.Headers:
        custom_headers = options.headers or {}
        headers_dict = self._merge_mappings(self._default_headers, custom_headers)
        headers = httpx.Headers(headers_dict)
        return headers

    def _merge_mappings(
        self,
        obj1: Mapping[str, Any],
        obj2: Mapping[str, Any],
    ) -> Dict[str, Any]:
        """Merge two mappings of the given type
        In cases with duplicate keys the second mapping takes precedence.
        """
        return {**obj1, **obj2}

    def is_closed(self) -> bool:
        return self._client.is_closed

    def close(self) -> None:
        """Close the underlying HTTPX client.

        The client will *not* be usable after this.
        """
        self._client.close()

    def __enter__(self: Any) -> Any:
        return self

    def __exit__(
        self,
        exc_type: Optional[BaseException],
        exc: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.close()

    def _build_request(self, options: Options) -> httpx.Request:
        headers = self._build_headers(options)
        params = options.params
        json_body = options.json_body
        request = self._client.build_request(
            method=options.method,
            url=options.url,
            headers=headers,
            params=params,
            json=json_body,
            timeout=options.timeout,
        )
        return request

    @overload
    def _request(
        self,
        *,
        options: Options,
        stream: Literal[False],
        cast_to: Type[ResponseT],
        stream_cls: Union[type[StreamT], None] = None,
    ) -> ResponseT:
        ...

    @overload
    def _request(
        self,
        *,
        options: Options,
        stream: Literal[True],
        cast_to: Type[ResponseT],
        stream_cls: Union[type[StreamT], None] = None,
    ) -> StreamT:
        ...

    @overload
    def _request(
        self,
        *,
        options: Options,
        stream: bool,
        cast_to: Type[ResponseT],
        stream_cls: Union[type[StreamT], None] = None,
    ) -> Union[ResponseT, StreamT]:
        ...

    def _request(
        self,
        *,
        options: Options,
        stream: bool,
        cast_to: Type[ResponseT],
        stream_cls: Union[type[StreamT], None] = None,
    ) -> Union[ResponseT, StreamT]:
        request = self._build_request(options)
        try:
            res = self._client.send(request, auth=self.custom_auth, stream=stream)
            res.raise_for_status()
        except httpx.HTTPStatusError as err:  # 4xx and 5xx errors
            # If the response is streamed then we need to explicitly read the response
            # to completion before attempting to access the response text.
            err.response.read()
            raise self._make_status_error_from_response(request, err.response) from None
        except httpx.TimeoutException as err:
            raise APITimeoutError(request=request) from err
        except Exception as err:
            raise APIConnectionError(request=request) from err
        if res.headers["content-type"] == "text/event-stream":
            stream_cls = stream_cls or cast(
                "Union[type[StreamT], None]", self._default_stream_cls
            )
            if stream_cls is None:
                raise MissingStreamClassError()
            stream_response = stream_cls(response=res)
            return stream_response
        response = cast(
            ResponseT,
            RubeusResponse.construct(**res.json(), raw_body=res.json()),
        )
        return response

    def _make_status_error_from_response(
        self,
        request: httpx.Request,
        response: httpx.Response,
    ) -> APIStatusError:
        err_text = response.text.strip()
        body = err_text

        try:
            body = json.loads(err_text)["error"]["message"]
            err_msg = f"Error code: {response.status_code} - {body}"
        except Exception:
            err_msg = err_text or f"Error code: {response.status_code}"

        return make_status_error(err_msg, body=body, request=request, response=response)



from typing import Optional, Union, List, Dict, Any, cast, overload, Literal

from rubeus.api_resources.base_client import APIClient
from .global_constants import DEFAULT_MAX_RETRIES, DEFAULT_TIMEOUT
from .utils import (
    ProviderTypes,
    RubeusCacheType,
    LLMBase,
    RubeusModes,
    Message,
    ProviderTypesLiteral,
    Body,
    RubeusResponse,
)

from .streaming import Stream

__all__ = ["Completions", "ChatCompletions"]


class APIResource:
    _client: APIClient

    def __init__(self, client: APIClient) -> None:
        self._client = client
        # self._get = client.get
        self._post = client.post
        # self._patch = client.patch
        # self._put = client.put
        # self._delete = client.delete
        # self._get_api_list = client.get_api_list


class Completions(APIResource):
    @overload
    def create(
        self,
        *,
        prompt: str = "",
        timeout: Union[float, None] = DEFAULT_TIMEOUT,
        max_retries: int = DEFAULT_MAX_RETRIES,
        provider: Union[ProviderTypes, ProviderTypesLiteral] = ProviderTypes.OPENAI,
        model: str = "gpt-3.5-turbo",
        api_key: str = "",
        temperature: float = 0.1,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        stop_sequences: Optional[List[str]] = None,
        max_tokens: Optional[int] = None,
        trace_id: Optional[str] = None,
        cache_status: Optional[RubeusCacheType] = None,
        cache: Optional[bool] = False,
        metadata: Optional[Dict[str, Any]] = None,
        weight: Optional[float] = 1.0,
        stream: Literal[True]
    ) -> Stream[RubeusResponse]:
        ...

    @overload
    def create(
        self,
        *,
        prompt: str = "",
        timeout: Union[float, None] = DEFAULT_TIMEOUT,
        max_retries: int = DEFAULT_MAX_RETRIES,
        provider: Union[ProviderTypes, ProviderTypesLiteral] = ProviderTypes.OPENAI,
        model: str = "gpt-3.5-turbo",
        api_key: str = "",
        temperature: float = 0.1,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        stop_sequences: Optional[List[str]] = None,
        max_tokens: Optional[int] = None,
        trace_id: Optional[str] = None,
        cache_status: Optional[RubeusCacheType] = None,
        cache: Optional[bool] = False,
        metadata: Optional[Dict[str, Any]] = None,
        weight: Optional[float] = 1.0,
        stream: Literal[False] = False
    ) -> RubeusResponse:
        ...

    @overload
    def create(
        self,
        *,
        prompt: str = "",
        timeout: Union[float, None] = DEFAULT_TIMEOUT,
        max_retries: int = DEFAULT_MAX_RETRIES,
        provider: Union[ProviderTypes, ProviderTypesLiteral] = ProviderTypes.OPENAI,
        model: str = "gpt-3.5-turbo",
        api_key: str = "",
        temperature: float = 0.1,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        stop_sequences: Optional[List[str]] = None,
        max_tokens: Optional[int] = None,
        trace_id: Optional[str] = None,
        cache_status: Optional[RubeusCacheType] = None,
        cache: Optional[bool] = False,
        metadata: Optional[Dict[str, Any]] = None,
        weight: Optional[float] = 1.0,
        stream: bool = False
    ) -> Union[RubeusResponse, Stream[RubeusResponse]]:
        ...

    def create(
        self,
        *,
        prompt: str = "",
        timeout: Union[float, None] = DEFAULT_TIMEOUT,
        max_retries: int = DEFAULT_MAX_RETRIES,
        provider: Union[ProviderTypes, ProviderTypesLiteral] = ProviderTypes.OPENAI,
        model: str = "gpt-3.5-turbo",
        api_key: str = "",
        temperature: float = 0.1,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        stop_sequences: Optional[List[str]] = None,
        max_tokens: Optional[int] = None,
        trace_id: Optional[str] = None,
        cache_status: Optional[RubeusCacheType] = None,
        cache: Optional[bool] = False,
        metadata: Optional[Dict[str, Any]] = None,
        weight: Optional[float] = 1.0,
        stream: bool = False
    ) -> Union[RubeusResponse, Stream[RubeusResponse]]:
        llm = Body(
            prompt=prompt,
            timeout=timeout,
            max_retries=max_retries,
            provider=provider,
            model=model,
            api_key=api_key,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            stop_sequences=stop_sequences,
            max_tokens=max_tokens,
            trace_id=trace_id,
            cache_status=cache_status,
            cache=cache,
            metadata=metadata,
            weight=weight,
        )
        return self._post(
            "/v1/complete",
            body=[llm],
            mode=RubeusModes.SINGLE.value,
            cast_to=RubeusResponse,
            stream_cls=Stream[RubeusResponse],
            stream=stream,
        )

    @overload
    def with_fallbacks(
        self, *, llms: List[LLMBase], stream: Literal[True]
    ) -> Stream[RubeusResponse]:
        ...

    @overload
    def with_fallbacks(
        self, *, llms: List[LLMBase], stream: Literal[False] = False
    ) -> RubeusResponse:
        ...

    @overload
    def with_fallbacks(
        self, *, llms: List[LLMBase], stream: bool = False
    ) -> Union[RubeusResponse, Stream[RubeusResponse]]:
        ...

    def with_fallbacks(
        self, llms: List[LLMBase], stream: bool = False
    ) -> Union[RubeusResponse, Stream[RubeusResponse]]:
        body = []
        for i in llms:
            body.append(cast(Body, i))
        return self._post(
            "/v1/chatComplete",
            body=body,
            mode=RubeusModes.FALLBACK,
            cast_to=RubeusResponse,
            stream_cls=Stream[RubeusResponse],
            stream=stream,
        )

    @overload
    def with_loadbalancing(
        self, llms: List[LLMBase], stream: Literal[True]
    ) -> Stream[RubeusResponse]:
        ...

    @overload
    def with_loadbalancing(
        self, llms: List[LLMBase], stream: Literal[False] = False
    ) -> RubeusResponse:
        ...

    @overload
    def with_loadbalancing(
        self, llms: List[LLMBase], stream: bool = False
    ) -> Union[RubeusResponse, Stream[RubeusResponse]]:
        ...

    def with_loadbalancing(
        self, llms: List[LLMBase], stream: bool = False
    ) -> Union[RubeusResponse, Stream[RubeusResponse]]:
        body = []
        for i in llms:
            body.append(cast(Body, i))
        return self._post(
            "/v1/chatComplete",
            body=body,
            mode=RubeusModes.LOADBALANCE,
            cast_to=RubeusResponse,
            stream_cls=Stream[RubeusResponse],
            stream=stream,
        )


class ChatCompletions(APIResource):
    @overload
    def create(
        self,
        *,
        messages: List[Message],
        provider: ProviderTypes = ProviderTypes.OPENAI,
        model: str = "gpt-3.5-turbo",
        api_key: str = "",
        timeout: Union[float, None] = DEFAULT_TIMEOUT,
        max_retries: int = DEFAULT_MAX_RETRIES,
        temperature: float = 0.1,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        stop_sequences: Optional[List[str]] = None,
        max_tokens: Optional[int] = None,
        trace_id: Optional[str] = "",
        cache_status: Optional[RubeusCacheType] = None,
        cache: Optional[bool] = False,
        metadata: Optional[Dict[str, Any]] = None,
        weight: Optional[float] = 1.0,
        stream: Literal[True]
        # retry_settings: Optional[RetrySettings] = None,
    ) -> Stream[RubeusResponse]:
        ...

    @overload
    def create(
        self,
        *,
        messages: List[Message],
        provider: ProviderTypes = ProviderTypes.OPENAI,
        model: str = "gpt-3.5-turbo",
        api_key: str = "",
        timeout: Union[float, None] = DEFAULT_TIMEOUT,
        max_retries: int = DEFAULT_MAX_RETRIES,
        temperature: float = 0.1,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        stop_sequences: Optional[List[str]] = None,
        max_tokens: Optional[int] = None,
        trace_id: Optional[str] = "",
        cache_status: Optional[RubeusCacheType] = None,
        cache: Optional[bool] = False,
        metadata: Optional[Dict[str, Any]] = None,
        weight: Optional[float] = 1.0,
        stream: Literal[False] = False
        # retry_settings: Optional[RetrySettings] = None,
    ) -> RubeusResponse:
        ...

    @overload
    def create(
        self,
        *,
        messages: List[Message],
        provider: ProviderTypes = ProviderTypes.OPENAI,
        model: str = "gpt-3.5-turbo",
        api_key: str = "",
        timeout: Union[float, None] = DEFAULT_TIMEOUT,
        max_retries: int = DEFAULT_MAX_RETRIES,
        temperature: float = 0.1,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        stop_sequences: Optional[List[str]] = None,
        max_tokens: Optional[int] = None,
        trace_id: Optional[str] = "",
        cache_status: Optional[RubeusCacheType] = None,
        cache: Optional[bool] = False,
        metadata: Optional[Dict[str, Any]] = None,
        weight: Optional[float] = 1.0,
        stream: bool = False
        # retry_settings: Optional[RetrySettings] = None,
    ) -> Union[RubeusResponse, Stream[RubeusResponse]]:
        ...

    def create(
        self,
        *,
        messages: List[Message],
        provider: ProviderTypes = ProviderTypes.OPENAI,
        model: str = "gpt-3.5-turbo",
        api_key: str = "",
        timeout: Union[float, None] = DEFAULT_TIMEOUT,
        max_retries: int = DEFAULT_MAX_RETRIES,
        temperature: float = 0.1,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        stop_sequences: Optional[List[str]] = None,
        max_tokens: Optional[int] = None,
        trace_id: Optional[str] = "",
        cache_status: Optional[RubeusCacheType] = None,
        cache: Optional[bool] = False,
        metadata: Optional[Dict[str, Any]] = None,
        weight: Optional[float] = 1.0,
        stream: bool = False
        # retry_settings: Optional[RetrySettings] = None,
    ) -> Union[RubeusResponse, Stream[RubeusResponse]]:
        llm = Body(
            messages=messages,
            timeout=timeout,
            max_retries=max_retries,
            provider=provider,
            model=model,
            api_key=api_key,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            stop_sequences=stop_sequences,
            max_tokens=max_tokens,
            trace_id=trace_id,
            cache_status=cache_status,
            cache=cache,
            metadata=metadata,
            weight=weight
            # retry_settings=retry_settings,
        )
        return self._post(
            "/v1/chatComplete",
            body=[llm],
            mode=RubeusModes.SINGLE.value,
            cast_to=RubeusResponse,
            stream_cls=Stream[RubeusResponse],
            stream=stream,
        )

    @overload
    def with_fallbacks(
        self, *, llms: List[LLMBase], stream: Literal[True]
    ) -> Stream[RubeusResponse]:
        ...

    @overload
    def with_fallbacks(
        self, *, llms: List[LLMBase], stream: Literal[False] = False
    ) -> RubeusResponse:
        ...

    @overload
    def with_fallbacks(
        self, *, llms: List[LLMBase], stream: bool = False
    ) -> Union[RubeusResponse, Stream[RubeusResponse]]:
        ...

    def with_fallbacks(
        self, *, llms: List[LLMBase], stream: bool = False
    ) -> Union[RubeusResponse, Stream[RubeusResponse]]:
        body = []
        for i in llms:
            body.append(cast(Body, i))
        return self._post(
            "/v1/chatComplete",
            body=body,
            mode=RubeusModes.FALLBACK,
            cast_to=RubeusResponse,
            stream_cls=Stream[RubeusResponse],
            stream=stream,
        )

    @overload
    def with_loadbalancing(
        self, llms: List[LLMBase], stream: Literal[True]
    ) -> Stream[RubeusResponse]:
        ...

    @overload
    def with_loadbalancing(
        self, llms: List[LLMBase], stream: Literal[False] = False
    ) -> RubeusResponse:
        ...

    @overload
    def with_loadbalancing(
        self, llms: List[LLMBase], stream: bool = False
    ) -> Union[RubeusResponse, Stream[RubeusResponse]]:
        ...

    def with_loadbalancing(
        self, llms: List[LLMBase], stream: bool = False
    ) -> Union[RubeusResponse, Stream[RubeusResponse]]:
        body = []
        for i in llms:
            body.append(cast(Body, i))
        return self._post(
            "/v1/chatComplete",
            body=body,
            mode=RubeusModes.LOADBALANCE,
            cast_to=RubeusResponse,
            stream_cls=Stream[RubeusResponse],
            stream=stream,
        )



"""Rubeus implementation."""
import os
from typing import Optional, Union, Any, Dict
from .global_constants import (
    MISSING_API_KEY_ERROR_MESSAGE,
    DEFAULT_MAX_RETRIES,
    DEFAULT_TIMEOUT,
)
from .utils import DefaultParams
from .base_client import APIClient
from . import apis

__all__ = ["Rubeus"]


class Rubeus(APIClient):
    completion: apis.Completions
    chat_completion: apis.ChatCompletions

    def __init__(
        self,
        *,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        timeout: Union[float, None] = DEFAULT_TIMEOUT,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Optional[Dict[str, Any]] = None,
        default_query: Optional[Dict[str, Any]] = None,
        default_params: Optional[DefaultParams] = None,
    ) -> None:
        if base_url is None:
            self.base_url = "https://api.portkey.ai"
        self.api_key = api_key or os.environ.get("PORTKEY_API_KEY") or ""
        if not self.api_key:
            raise ValueError(MISSING_API_KEY_ERROR_MESSAGE)

        self.default_params = {} if default_params is None else default_params.dict()
        self.timeout = timeout
        self.max_retries = max_retries
        self.default_headers = default_headers or {}
        self.default_query = default_query or {}
        super().__init__(
            base_url=self.base_url,
            api_key=self.api_key,
            timeout=self.timeout,
            max_retries=self.max_retries,
            custom_headers=self.default_headers,
            custom_query=self.default_query,
            custom_params=self.default_params,
        )
        self.completion = apis.Completions(self)
        self.chat_completion = apis.ChatCompletions(self)

    def __setitem__(self, key: str, value: Any) -> None:
        setattr(self, key, value)



""""""
from .client import Rubeus
from .utils import (
    RubeusModes,
    RubeusModesLiteral,
    LLMBase,
    ProviderTypes,
    ProviderTypesLiteral,
    RubeusCacheType,
    RubeusCacheLiteral,
    Message,
    RubeusResponse,
    DefaultParams,
)

from rubeus.version import VERSION

__version__ = VERSION
__all__ = [
    "Rubeus",
    "LLMBase",
    "RubeusModes",
    "RubeusResponse",
    "RubeusModesLiteral",
    "ProviderTypes",
    "ProviderTypesLiteral",
    "RubeusCacheType",
    "RubeusCacheLiteral",
    "Message",
    "DefaultParams",
]
